// Sequitur version 1.00

#include <unistd.h>
#include <fstream.h>
#include <sys/times.h>
#include <limits.h>
#include "classes.h"

rules *S;

int chars = 0;
int num_rules = 0;
int num_symbols = 0;
int K = 1;

int compress = 0, do_uncompress = 0, do_print = 0, reproduce = 0, max_symbols = 0,
  quiet = 0, phind = 0, numbers = 0, print_rule_freq = 0, print_rule_usage = 0, delimiter = -1; 

char *delimiter_string = 0;

void uncompress(), print(), number(), forget(symbols *s), 
  forget_print(symbols *s);
void start_compress(), end_compress(), stop_forgetting();
ofstream *rule_S = 0;

char *help = "\n\
usage: sequitur -cdpqrtTuz -k <K> -e <delimiter> -f <max symbols>\n\n\
-p    print grammar at end\n\
-d    treat input as symbol numbers, one per line\n\
-c    compress\n\
-u    uncompress\n\
-q    quiet: suppress progress numbers on stderr\n\
-r    reproduce full expansion of rules after each rule\n\
-t    print rule usage in the grammar after each rule\n\
-T    print rule usage in the input after each rule\n\
-z    put rule S in file called S, other rules on stdout as usual\n\
-k    set K, the minmum number of times a digram must occur to form rule\n\
      (default 2)\n\
-e    set the delimiter symbol. Rules will not be formed across (i.e. \n\
      including) delimiters. If with -d, 0-9 are treated as numbers\n\
-f    set maximum symbols in grammar (memory limit). Grammar/compressed output\n\
      will be generated once the grammar reaches this size\n\
";

main(int argc, char **argv)
{
  extern char *optarg;
  extern int optind, opterr;
  int c;

  while ((c = getopt(argc, argv, "cuk:prf:qzdtTe:h")) != -1)
    switch (c) {
    case 'h': cerr << help; exit(2); break;
    case 't': print_rule_freq = 1; break;
    case 'T': print_rule_usage = 1; break;
    case 'd': numbers = 1; break;
    case 'c': compress = 1; break;
    case 'u': do_uncompress = 1; break;
    case 'p': do_print = 1; break;
    case 'r': reproduce = 1; break;
    case 'q': quiet = 1; break;
    case 'z': phind = 1; break;
    case 'e': delimiter_string = optarg; break;
    case 'f': max_symbols = atoi(optarg); break;
    case 'k': K = atoi(optarg) - 1; break;
    }

  if (delimiter_string)
    delimiter = numbers ? atoi(delimiter_string) : delimiter_string[0];

  if (compress) start_compress();

  if (do_uncompress) { 
    uncompress();
    exit(0);
  }

  if (phind) current_rule = 1;

  S = new rules;

  int i;

  if (numbers) cin >> i;
  else i = cin.get(); 

  S->last()->insert_after(new symbols(i));

  while (1) {
    if (++ chars % 100000 == 0 && !quiet) {
      static int last_time = 0;
      struct tms buffer;
      extern int collisions, seeks, occupancy;//, occ[100];
      times(&buffer);

      cerr << chars << '\t' << (buffer.tms_utime - last_time) << endl;
      last_time = buffer.tms_utime;
    }
    if (numbers) cin >> i;
    else i = cin.get(); 

    if (cin.eof()) break;
    S->last()->insert_after(new symbols(i));
    S->last()->prev()->check();
    if (max_symbols && num_symbols > max_symbols)
      if (compress) forget(S->first());
      else if (phind) forget_print(S->first());
  }

  
  void calculate_rule_usage(rules *r);
  if (print_rule_usage) calculate_rule_usage(S);

  if (max_symbols || compress || phind) {
    if (compress) stop_forgetting();
    while (S->first()->next() != S->first()) 
      if (compress) forget(S->first());
      else if (phind) forget_print(S->first());
  }

  if (compress) end_compress();
  
  if (do_print) {
     number();
     print();
  }
}

// print the rules out

rules **R1;
int Ri;

void print()
{
  for (int i = 0; i < Ri; i ++) {
    if (i > 0) cout << i << " -> ";
    for (symbols *p = R1[i]->first(); !p->is_guard(); p = p->next())
      if (p->nt()) cout << p->rule()->index() << ' ';
      else cout << *p << ' ';
    if (i > 0 && print_rule_freq) cout << '\t' << R1[i]->freq();
    if (i > 0 && print_rule_usage) cout << "\t(" << R1[i]->usage() << ")";
    if (reproduce && i > 0) {
      cout << '\t';
      R1[i]->reproduce();
    }
    cout << endl;
  }
  
  cout << (num_symbols - Ri) << " symbols, " << Ri << " rules " <<
    (num_symbols * (sizeof(symbols) + 4) + Ri * sizeof(rules)) << " total space\n";
  
}

void number()
{
  R1 = (rules **) malloc(sizeof(rules *) * num_rules);
  memset(R1, 0, sizeof(rules *) * num_rules);
  R1[0] = S;
  Ri = 1;

  for (int i = 0; i < Ri; i ++)
    for (symbols *p = R1[i]->first(); !p->is_guard(); p = p->next())
      if (p->nt() && R1[p->rule()->index()] != p->rule()) {
	p->rule()->index(Ri);
	R1[Ri ++] = p->rule();
      }
}


void forget_print(symbols *s) 
{
  if (rule_S == 0) rule_S = new ofstream("S");

  if (s->nt()) {
    rules *r = s->rule();
    delete s;

    if (r->freq() == 0) { 
      *rule_S << r->index();
      while (r->first()->next() != r->first()) delete r->first();
      delete r;
    }
    else {
      if (r->index() == 0) r->output();
      *rule_S << r->index();
    }

  }
  else {
    *rule_S << *s;
    delete s;
  }

  *rule_S << ' ';
}


void calculate_rule_usage(rules *r)
{
  for (symbols *p = r->first(); !p->is_guard(); p = p->next()) {
    if (p->nt()) {
      p->rule()->usage(1);
      calculate_rule_usage(p->rule());
    }
  }
}

